@@ -32,6 +32,16 @@ class OEMHandler(object):
     def get_video_launchdata(self):
         return {}
 
+    def get_description(self):
+        """Get a description of descriptive attributes of a node.
+
+        Height describes, in U how tall the system is, and slot is 0 if
+        not a blade type server, and slot if it is.
+
+        :return: dictionary with 'height' and 'slot' members
+        """
+        return {}
+
     def process_event(self, event, ipmicmd, seldata):
         """Modify an event according with OEM understanding.
 
@@ -46,6 +56,10 @@ class OEMHandler(object):
         if evdata[0] & 0b110000 == 0b100000:
             event['oem_byte3'] = evdata[2]
 
+    def clear_system_configuration(self):
+        raise exc.UnsupportedFunctionality(
+            'Clearing system configuration not implemented for this platform')
+
     def get_oem_inventory_descriptions(self):
         """Get descriptions of available additional inventory items
 
@@ -75,6 +89,18 @@ class OEMHandler(object):
         """
         return ()
 
+    def get_diagnostic_data(self, savefile, progress=None):
+        """Download diagnostic data about target to a file
+
+        This should be a payload that the vendor's support team can use
+        to do diagnostics.
+        :param savefile: File object or filename to save to
+        :param progress: Callback to be informed about progress
+        :return:
+        """
+        raise exc.UnsupportedFunctionality(
+            'Do not know how to get diagnostic data for this platform')
+
     def get_sensor_data(self):
         """Get OEM sensor data
 
@@ -141,7 +167,7 @@ class OEMHandler(object):
         """
         return None
 
-    def process_fru(self, fru):
+    def process_fru(self, fru, name=None):
         """Modify a fru entry with OEM understanding.
 
         Given a fru, clarify 'extra' fields according to OEM rules and
@@ -159,12 +185,18 @@ class OEMHandler(object):
         fru['oem_parser'] = None
         return fru
 
-    def get_oem_firmware(self, bmcver):
+    def get_oem_firmware(self, bmcver, components):
         """Get Firmware information.
         """
         # Here the bmc version is passed into the OEM handler, to allow
         # the handler to enrich the data. For the generic case, just
         # provide the generic BMC version, which is all that is possible
+        # Additionally, components may be provided for an advisory guide
+        # on interesting firmware.  The OEM library is permitted to return
+        # more than requested, and it is the responsibility of the calling
+        # code to know whether it cares or not.  The main purpose of the
+        # components argument is to indicate when certain performance
+        # optimizations can be performed.
         yield ('BMC Version', {'version': bmcver})
 
     def get_oem_capping_enabled(self):
@@ -259,6 +291,35 @@ class OEMHandler(object):
         """
         raise exc.UnsupportedFunctionality()
 
+    def get_health(self, summary):
+        """Provide an alternative or augmented health assessment
+
+        An OEM handler can preprocess the summary and extend it with OEM
+        specific data, and then return to let generic processing occur.
+        It can also raise the pyghmi exception BypassGenericBehavior to
+        suppress the standards based routine, for enhanced performance.
+
+        :param summary: The health summary as prepared by the generic function
+        :return: Nothing, modifies the summary object
+        """
+        return []
+
+    def set_hostname(self, hostname):
+        """OEM specific hook to specify name information
+
+        """
+        raise exc.UnsupportedFunctionality()
+
+    def get_hostname(self):
+        """OEM specific hook to specify name information
+
+        """
+        raise exc.UnsupportedFunctionality()
+
+    def set_user_access(self, uid, channel, callback, link_auth, ipmi_msg,
+                        privilege_level):
+        return  # Nothing to do
+
     def set_alert_ipv6_destination(self, ip, destination, channel):
         """Set an IPv6 alert destination
 
@@ -273,3 +334,36 @@ class OEMHandler(object):
         :returns True if standard parameter set should be suppressed
         """
         return False
+
+    def get_system_configuration(self, hideadvanced):
+        """Retrieve system configuration
+
+        This returns a dictionary of settings names to dictionaries including
+        'current', 'default' and 'possible' values as well as 'help'
+
+        :param hideadvanced: Whether to hide 'advanced' settings that most
+                             users should not need.  Defaults to True.
+        """
+        return {}
+
+    def set_system_configuration(self, changeset):
+        """Apply a changeset to system configuration
+
+        Takes a key value pair and applies it against the system configuration
+        """
+        raise exc.UnsupportedFunctionality()
+
+    def get_licenses(self):
+        raise exc.UnsupportedFunctionality()
+
+    def delete_license(self, name):
+        raise exc.UnsupportedFunctionality()
+
+    def save_licenses(self, directory):
+        raise exc.UnsupportedFunctionality()
+
+    def apply_license(self, filename, progress=None):
+        raise exc.UnsupportedFunctionality()
+
+    def get_user_expiration(self, uid):
+        return None
