@@ -326,10 +326,19 @@ class OEMHandler(generic.OEMHandler):
             return True
         return None
 
+    def set_user_access(self, uid, channel, callback, link_auth, ipmi_msg,
+                        privilege_level):
+        if self.is_fpc and  self._fpc_variant == 2:
+            self.smmhandler.set_user_priv(uid, privilege_level)
+
     @property
     def is_fpc(self):
         """True if the target is a Lenovo nextscale fan power controller
         """
+        if self.has_imm or self.has_xcc:
+            return None
+        if self._fpc_variant is not None:
+            return self._fpc_variant
         fpc_ids = ((19046, 32, 1063), (20301, 32, 462))
         smm_id = (19046, 32, 1180)
         currid = (self.oemid['manufacturer_id'], self.oemid['device_id'],
@@ -381,29 +390,29 @@ class OEMHandler(generic.OEMHandler):
                 yield inv
 
     def get_sensor_data(self):
-        if self.is_fpc:
-            for name in nextscale.get_sensor_names(self._fpc_variant):
-                yield nextscale.get_sensor_reading(name, self.ipmicmd,
-                                                   self._fpc_variant)
-        elif self.has_imm:
+        if self.has_imm:
             for name in self.immhandler.get_oem_sensor_names(self.ipmicmd):
                 yield self.immhandler.get_oem_sensor_reading(name,
                                                              self.ipmicmd)
+        elif self.is_fpc:
+            for name in nextscale.get_sensor_names(self._fpc_variant):
+                yield nextscale.get_sensor_reading(name, self.ipmicmd,
+                                                   self._fpc_variant)
 
     def get_sensor_descriptions(self):
-        if self.is_fpc:
-            return nextscale.get_sensor_descriptions(self._fpc_variant)
-        elif self.has_imm:
+        if self.has_imm:
             return self.immhandler.get_oem_sensor_descriptions(self.ipmicmd)
+        elif self.is_fpc:
+            return nextscale.get_sensor_descriptions(self._fpc_variant)
         return ()
 
     def get_sensor_reading(self, sensorname):
-        if self.is_fpc:
-            return nextscale.get_sensor_reading(sensorname, self.ipmicmd,
-                                                self._fpc_variant)
-        elif self.has_imm:
+        if self.has_imm:
             return self.immhandler.get_oem_sensor_reading(sensorname,
                                                           self.ipmicmd)
+        elif self.is_fpc:
+            return nextscale.get_sensor_reading(sensorname, self.ipmicmd,
+                                                self._fpc_variant)
         return ()
 
     def get_inventory_of_component(self, component):
@@ -483,7 +492,7 @@ class OEMHandler(generic.OEMHandler):
         else:
             raise pygexc.UnsupportedFunctionality()
 
-    def process_fru(self, fru):
+    def process_fru(self, fru, name=None):
         if fru is None:
             return fru
         if self.has_tsm:
@@ -547,6 +556,8 @@ class OEMHandler(generic.OEMHandler):
                     idx = idx + 6
             except (AttributeError, KeyError, IndexError):
                 pass
+            if self.has_xcc and name and name.startswith('PSU '):
+                self.immhandler.augment_psu_info(fru, name)
             return fru
         elif self.is_fpc == 2:  # SMM variant
             fru['oem_parser'] = 'lenovo'
@@ -595,17 +606,23 @@ class OEMHandler(generic.OEMHandler):
         self._hasimm = (rdata[1] & 1 == 1) or (rdata[1] & 16 == 16)
         return self._hasimm
 
-    def get_oem_firmware(self, bmcver):
+    def get_oem_firmware(self, bmcver, components):
         if self.has_tsm:
             command = firmware.get_categories()["firmware"]
             rsp = self.ipmicmd.xraw_command(**command["command"])
             return command["parser"](rsp["data"])
         elif self.has_imm:
-            return self.immhandler.get_firmware_inventory(bmcver)
+            return self.immhandler.get_firmware_inventory(bmcver, components)
         elif self.is_fpc:
             return nextscale.get_fpc_firmware(bmcver, self.ipmicmd,
                                               self._fpc_variant)
-        return super(OEMHandler, self).get_oem_firmware(bmcver)
+        return super(OEMHandler, self).get_oem_firmware(bmcver, components)
+
+    def get_diagnostic_data(self, savefile, progress):
+        if self.has_xcc:
+            return self.immhandler.get_diagnostic_data(savefile, progress)
+        if self.is_fpc:
+            return self.smmhandler.get_diagnostic_data(savefile, progress)
 
     def get_oem_capping_enabled(self):
         if self.has_tsm:
@@ -652,6 +669,8 @@ class OEMHandler(generic.OEMHandler):
                                                 data=(4, i))
                 name += rsp['data'][:]
             return name.rstrip('\x00')
+        elif self.is_fpc:
+            return self.smmhandler.get_domain()
 
     def set_oem_domain_name(self, name):
         if self.has_tsm:
@@ -669,6 +688,22 @@ class OEMHandler(generic.OEMHandler):
 
             self._restart_dns()
             return
+        elif self.is_fpc:
+            self.smmhandler.set_domain(name)
+
+    def set_hostname(self, hostname):
+        if self.has_xcc:
+            return self.immhandler.set_hostname(hostname)
+        elif self.is_fpc:
+            return self.smmhandler.set_hostname(hostname)
+        return super(OEMHandler, self).set_hostname(hostname)
+
+    def get_hostname(self):
+        if self.has_xcc:
+            return self.immhandler.get_hostname()
+        elif self.is_fpc:
+            return self.smmhandler.get_hostname()
+        return super(OEMHandler, self).get_hostname()
 
     """ Gets a remote console launcher for a Lenovo ThinkServer.
 
@@ -896,6 +931,28 @@ class OEMHandler(generic.OEMHandler):
                                                        progress=progress,
                                                        bank=bank)
 
+    def get_description(self):
+        if self.has_xcc:
+            return self.immhandler.get_description()
+        if self.is_fpc:
+            return {'height': self._fpc_variant, 'slot': 0}
+        return super(OEMHandler, self).get_description()
+
+    def get_system_configuration(self, hideadvanced):
+        if self.has_imm or self.has_xcc:
+            return self.immhandler.get_system_configuration(hideadvanced)
+        return super(OEMHandler, self).get_system_configuration(hideadvanced)
+
+    def set_system_configuration(self, changeset):
+        if self.has_imm or self.has_xcc:
+            return self.immhandler.set_system_configuration(changeset)
+        return super(OEMHandler, self).set_system_configuration(changeset)
+
+    def clear_system_configuration(self):
+        if self.has_xcc:
+            return self.immhandler.clear_system_configuration()
+        return super(OEMHandler, self).clear_system_configuration()
+
     def detach_remote_media(self):
         if self.has_imm:
             self.immhandler.detach_remote_media()
@@ -905,11 +962,41 @@ class OEMHandler(generic.OEMHandler):
             self.ipmicmd.xraw_command(netfn=0x32, command=0x9f, data=(8, 11))
 
     def upload_media(self, filename, progress):
-        if self.has_xcc:
+        if self.has_xcc or self.has_imm:
             return self.immhandler.upload_media(filename, progress)
         return super(OEMHandler, self).upload_media(filename, progress)
 
     def list_media(self):
-        if self.has_xcc:
+        if self.has_xcc or self.has_imm:
             return self.immhandler.list_media()
         return super(OEMHandler, self).list_media()
+
+    def get_health(self, summary):
+        if self.has_xcc:
+            return self.immhandler.get_health(summary)
+        return super(OEMHandler, self).get_health(summary)
+
+    def get_licenses(self):
+        if self.has_xcc:
+            return self.immhandler.get_licenses()
+        return super(OEMHandler, self).get_licenses()
+
+    def get_user_expiration(self, uid):
+        if self.has_xcc:
+            return self.immhandler.get_user_expiration(uid)
+        return None
+
+    def delete_license(self, name):
+        if self.has_xcc:
+            return self.immhandler.delete_license(name)
+        return super(OEMHandler, self).delete_license(name)
+
+    def save_licenses(self, directory):
+        if self.has_xcc:
+            return self.immhandler.save_licenses(directory)
+        return super(OEMHandler, self).save_licenses(directory)
+
+    def apply_license(self, filename, progress=None):
+        if self.has_xcc:
+            return self.immhandler.apply_license(filename, progress)
+        return super(OEMHandler, self).apply_license(filename, progress)
