@@ -1148,7 +1148,7 @@ sensor_type_offsets = {
     0x1e: {  # boot error
         0: {
             'desc': 'No bootable media',
-            'severity': const.Health.Failed,
+            'severity': const.Health.Ok,
             'deassertion_desc': 'No bootable media deasserted',
             'deassertion_severity': const.Health.Ok,
         },
@@ -1285,7 +1285,7 @@ sensor_type_offsets = {
             'deassertion_severity': const.Health.Ok,
         },
     },
-    0x21:  {  # slot/connector
+    0x21: {  # slot/connector
         0x0: {
             'desc': 'Fault',
             'severity': const.Health.Critical,
@@ -1877,3 +1877,11 @@ ipmi_completion_codes = {
     0xff: "Unspecified",
     0xffff: "Timeout",  # not ipmi, but used internally
 }
+
+priv_levels = {
+     'callback': 1,
+     'user': 2,
+     'operator': 3,
+     'admin': 4,
+     'oem': 5,
+}
\ No newline at end of file
